import { constants } from 'http2';
import * as v from 'valibot';
import {HttpMiddlewareError, PlainResponse, Middleware} from '../server';

export const handleGetCollection: Middleware = async (req) => {
	const { query, resource, backend } = req;

	let data: v.Output<typeof resource.schema>[];
	let totalItemCount: number | undefined;
	try {
		// TODO querying mechanism
		data = await resource.dataSource.getMultiple(query); // TODO paginated responses per resource
		if (backend!.showTotalItemCountOnGetCollection && typeof resource.dataSource.getTotalCount === 'function') {
			totalItemCount = await resource.dataSource.getTotalCount(query);
		}
	} catch (cause) {
		throw new HttpMiddlewareError(
			'unableToFetchResourceCollection',
			{
				cause,
				statusCode: constants.HTTP_STATUS_INTERNAL_SERVER_ERROR,
			}
		);
	}

	const headers: Record<string, string> = {};
	if (typeof totalItemCount !== 'undefined') {
		headers['X-Resource-Total-Item-Count'] = totalItemCount.toString();
	}

	return new PlainResponse({
		headers,
		statusCode: constants.HTTP_STATUS_OK,
		statusMessage: 'resourceCollectionFetched',
		body: data,
	});
};

export const handleGetItem: Middleware = async (req) => {
	const { resource, resourceId } = req;

	if (typeof resourceId === 'undefined' || resourceId.trim().length < 1) {
		throw new HttpMiddlewareError(
			'resourceIdNotGiven',
			{
				statusCode: constants.HTTP_STATUS_BAD_REQUEST,
			}
		);
	}

	let data: v.Output<typeof resource.schema> | null = null;
	try {
		data = await resource.dataSource.getById(resourceId);
	} catch (cause) {
		throw new HttpMiddlewareError(
			'unableToFetchResource',
			{
				cause,
				statusCode: constants.HTTP_STATUS_INTERNAL_SERVER_ERROR,
			}
		);
	}

	if (typeof data !== 'undefined' && data !== null) {
		return new PlainResponse({
			statusCode: constants.HTTP_STATUS_OK,
			statusMessage: 'resourceFetched',
			body: data
		});
	}

	throw new HttpMiddlewareError(
		'resourceNotFound',
		{
			statusCode: constants.HTTP_STATUS_NOT_FOUND,
		}
	);
};

export const handleDeleteItem: Middleware = async (req) => {
	const { resource, resourceId, backend } = req;

	if (typeof resourceId === 'undefined' || resourceId.trim().length < 1) {
		throw new HttpMiddlewareError(
			'resourceIdNotGiven',
			{
				statusCode: constants.HTTP_STATUS_BAD_REQUEST,
			}
		);
	}

	let existing: unknown | null;
	try {
		existing = await resource.dataSource.getById(resourceId);
	} catch (cause) {
		throw new HttpMiddlewareError('unableToFetchResource', {
			cause,
			statusCode: constants.HTTP_STATUS_INTERNAL_SERVER_ERROR
		});
	}

	if (!existing && backend!.throwsErrorOnDeletingNotFound) {
		throw new HttpMiddlewareError('deleteNonExistingResource', {
			statusCode: constants.HTTP_STATUS_NOT_FOUND
		});
	}

	try {
		if (existing) {
			await resource.dataSource.delete(resourceId);
		}
	} catch (cause) {
		throw new HttpMiddlewareError('unableToDeleteResource', {
			cause,
			statusCode: constants.HTTP_STATUS_INTERNAL_SERVER_ERROR
		})
	}

	return new PlainResponse({
		statusCode: constants.HTTP_STATUS_NO_CONTENT,
		statusMessage: 'resourceDeleted',
	});
};

export const handlePatchItem: Middleware = async (req) => {
	const { resource, resourceId, body } = req;

	if (typeof resourceId === 'undefined' || resourceId.trim().length < 1) {
		throw new HttpMiddlewareError(
			'resourceIdNotGiven',
			{
				statusCode: constants.HTTP_STATUS_BAD_REQUEST,
			}
		);
	}

	let existing: unknown | null;
	try {
		existing = await resource.dataSource.getById(resourceId!);
	} catch (cause) {
		throw new HttpMiddlewareError('unableToFetchResource', {
			cause,
			statusCode: constants.HTTP_STATUS_INTERNAL_SERVER_ERROR,
		});
	}

	if (!existing) {
		throw new HttpMiddlewareError('patchNonExistingResource', {
			statusCode: constants.HTTP_STATUS_NOT_FOUND,
		});
	}

	let newObject: v.Output<typeof resource.schema> | null;
	try {
		newObject = await resource.dataSource.patch(resourceId!, body as object);
	} catch (cause) {
		throw new HttpMiddlewareError('unableToPatchResource', {
			cause,
			statusCode: constants.HTTP_STATUS_INTERNAL_SERVER_ERROR
		});
	}

	return new PlainResponse({
		statusCode: constants.HTTP_STATUS_OK,
		statusMessage: 'resourcePatched',
		body: newObject,
	});
};

export const handleCreateItem: Middleware = async (req) => {
	const { resource, body, backend, basePath } = req;

	const idAttrRaw = resource.state.shared.get('idAttr');
	if (typeof idAttrRaw === 'undefined') {
		throw new HttpMiddlewareError('unableToGenerateIdFromResourceDataSource', {
			statusCode: constants.HTTP_STATUS_INTERNAL_SERVER_ERROR,
		});
	}
	const idAttr = idAttrRaw as string;

	let newId;
	let params: v.Output<typeof resource.schema>;
	try {
		newId = await resource.dataSource.newId();
		params = { ...body as Record<string, unknown> };
		params[idAttr] = newId;
	} catch (cause) {
		throw new HttpMiddlewareError('unableToGenerateIdFromResourceDataSource', {
			cause,
			statusCode: constants.HTTP_STATUS_INTERNAL_SERVER_ERROR,
		});
	}

	let newObject;
	let totalItemCount: number | undefined;
	try {
		newObject = await resource.dataSource.create(params);
		if (backend!.showTotalItemCountOnCreateItem && typeof resource.dataSource.getTotalCount === 'function') {
			totalItemCount = await resource.dataSource.getTotalCount();
		}
	} catch (cause) {
		throw new HttpMiddlewareError('unableToCreateResource', {
			cause,
			statusCode: constants.HTTP_STATUS_INTERNAL_SERVER_ERROR,
		});
	}

	const location = `${basePath}/${resource.state.routeName}/${newId}`;

	if (typeof totalItemCount !== 'undefined') {
		return new PlainResponse({
			statusCode: constants.HTTP_STATUS_CREATED,
			headers: {
				'Location': location,
				'X-Resource-Total-Item-Count': totalItemCount.toString()
			},
			body: newObject,
			statusMessage: 'resourceCreated'
		});
	}

	return new PlainResponse({
		statusCode: constants.HTTP_STATUS_CREATED,
		body: newObject,
		headers: {
			'Location': location,
		},
		statusMessage: 'resourceCreated'
	});
}

export const handleEmplaceItem: Middleware =  async (req) => {
	const { resource, resourceId, basePath, body, backend } = req;

	const idAttrRaw = resource.state.shared.get('idAttr');
	if (typeof idAttrRaw === 'undefined') {
		throw new HttpMiddlewareError('unableToGenerateIdFromResourceDataSource', {
			statusCode: constants.HTTP_STATUS_INTERNAL_SERVER_ERROR,
		});
	}
	const idAttr = idAttrRaw as string;

	let newObject: v.Output<typeof resource.schema>;
	let isCreated: boolean;
	try {
		const params = { ...body as Record<string, unknown> };
		params[idAttr] = resourceId;
		[newObject, isCreated] = await resource.dataSource.emplace(resourceId!, params);
	} catch (cause) {
		throw new HttpMiddlewareError('unableToEmplaceResource', {
			cause,
			statusCode: constants.HTTP_STATUS_INTERNAL_SERVER_ERROR,
		});
	}

	const headers: Record<string, string> = {};
	let totalItemCount: number | undefined;
	if (backend!.showTotalItemCountOnCreateItem && typeof resource.dataSource.getTotalCount === 'function') {
		totalItemCount = await resource.dataSource.getTotalCount();
	}
	if (isCreated) {
		headers['Location'] = `${basePath}/${resource.state.routeName}/${resourceId}`;
		if (typeof totalItemCount !== 'undefined') {
			headers['X-Resource-Total-Item-Count'] = totalItemCount.toString();
		}
	}

	return new PlainResponse({
		statusCode: isCreated ? constants.HTTP_STATUS_CREATED : constants.HTTP_STATUS_OK,
		headers,
		statusMessage: (
			isCreated
				? 'resourceCreated'
				: 'resourceReplaced'
		),
		body: newObject
	});
}
